#include "stdafx.h"
#include "PlayerLive.h"

#include "CurlHelper.h"
#include "tinyxml2.h"
#include <time.h>



PlayerLive::PlayerLive(NPVRSession* session, int channelOID, PlayerDelegate* playerDelegate, char* url, SDL_Renderer* renderer, SDL_Texture* videoTexture, SDL_mutex* mutex, bool useServerSkip, int forcedDuration, int64_t forceSize, bool useGPU)
	: Player(playerDelegate, url, renderer, videoTexture, mutex, useServerSkip, forcedDuration, forceSize, useGPU)
{
	this->session = session;
	this->channelOID = channelOID;
	this->lastRenewTime = time(NULL);

	lastKnownPosition = 0;
	lastKnownPositionTimestamp = SDL_GetTicks();

	session->GetBaseURL(baseURL, sizeof(baseURL));
	session->GetSID(sid, sizeof(sid));

	sprintf(clientID, "webclient-%s", sid);		
}

PlayerLive::~PlayerLive(void)
{
	StopStream();
}

void PlayerLive::Seek(int ms, int duration)
{
	int64_t streamLength = 1;
	int64_t streamDurationMS = 1;

	playerDelegate->PlaybackPosition(ms / 1000);

	// we need to check the state of the live tv buffer
	char url[2048];
	sprintf(url, "%s/service?method=channel.stream.info&format=xml&sid=%s", baseURL, sid);	
	char contentType[64] = "\0";
	struct MemoryStruct activityChunk;
	activityChunk.memory = (unsigned char*)malloc(1);
	activityChunk.size = 0;    
	DownloadURL(url, &activityChunk, contentType);

	// load xml dom	
	tinyxml2::XMLDocument doc;
	doc.Parse((const char*)activityChunk.memory);
	tinyxml2::XMLElement* mapNode = doc.FirstChildElement("map");
	if (mapNode != NULL)
	{
		tinyxml2::XMLElement* streamLengthNode = mapNode->FirstChildElement("stream_length");
		if (streamLengthNode != NULL)
		{
			streamLength = atoll(streamLengthNode->GetText());
		}

		tinyxml2::XMLElement* streamDurationNode = mapNode->FirstChildElement("stream_duration");
		if (streamDurationNode != NULL)
		{
			streamDurationMS = atoll(streamDurationNode->GetText());
		}
	}

	// check bounds
	if (ms < 0)
		ms = 0;
	else if (ms > streamDurationMS)
		ms = (streamDurationMS - 1000);

	// calc likely offset in live tv buffer
	int64_t startOffset = ((streamLength / streamDurationMS) * ms);
	startOffset = startOffset / 188;
	startOffset = startOffset * 188;

	// start playback of new url
	sprintf(url, "%s/live?channel_id=%d&sid=%s&seek=%d-", baseURL, channelOID, sid, startOffset);
	PlayStream(url);

	lastKnownPosition = ms;
	lastKnownPositionTimestamp = SDL_GetTicks();

	playerDelegate->PlaybackPosition(ms / 1000);
}


bool PlayerLive::StartStream(int64_t offset)
{
	char url[2048];
	sprintf(url, "%s/service?method=channel.stream.start&channel_id=%d&format=xml&sid=%s", baseURL, channelOID, sid);

	// check if any action was requested
	char contentType[64] = "\0";
	struct MemoryStruct activityChunk;
	activityChunk.memory = (unsigned char*)malloc(1);  /* will be grown as needed by the realloc above */
	activityChunk.size = 0;    /* no data at this point */
	DownloadURL(url, &activityChunk, contentType, 10000);

	this->lastRenewTime = time(NULL);

	return true;
}

void PlayerLive::StopStream()
{
	char url[2048];
	sprintf(url, "%s/service?method=channel.stream.stop&format=xml&sid=%s", baseURL, sid);

	// check if any action was requested
	char contentType[64] = "\0";
	struct MemoryStruct activityChunk;
	activityChunk.memory = (unsigned char*)malloc(1);  /* will be grown as needed by the realloc above */
	activityChunk.size = 0;    /* no data at this point */
	DownloadURL(url, &activityChunk, contentType);
}

void PlayerLive::RenewLease()
{
	printf("renewing lease\n");

	char url[2048];
	sprintf(url, "%s/service?method=channel.transcode.lease&format=xml&sid=%s", baseURL, sid);

	// check if any action was requested
	char contentType[64] = "\0";
	struct MemoryStruct activityChunk;
	activityChunk.memory = (unsigned char*)malloc(1);  /* will be grown as needed by the realloc above */
	activityChunk.size = 0;    /* no data at this point */
	DownloadURL(url, &activityChunk, contentType);

	this->lastRenewTime = time(NULL);
}

void PlayerLive::Tick()
{
	if (this->lastRenewTime + 5 < time(NULL))
	{
		RenewLease();
	}

	Player::Tick();
}